import multiprocessing
import logging
import time
import os

from logging.handlers import RotatingFileHandler
from logging.handlers import TimedRotatingFileHandler


class LockingRotatingFileHandler(RotatingFileHandler):
    _lock = multiprocessing.Lock()

    def __init__(self, filename, mode='a', maxBytes=0, backupCount=0,
                 encoding=None, delay=False, errors=None):
        super().__init__(filename, mode, maxBytes, backupCount,
                         encoding, delay, errors)

    def emit(self, record):
        try:
            if self.shouldRollover(record):
                with LockingRotatingFileHandler._lock:
                    self.close()
                    self.stream = self._open()
                    if self.shouldRollover(record):
                        self.doRollover()
            logging.FileHandler.emit(self, record)
        except Exception:
            self.handleError(record)


class LockingTimedRotatingFileHandler(TimedRotatingFileHandler):
    _lock = multiprocessing.Lock()

    def __init__(self, filename, when='h', interval=1, backupCount=0,
                 encoding=None, delay=False, utc=False, atTime=None,
                 errors=None):
        super().__init__(filename, when, interval, backupCount,
                         encoding, delay, utc, atTime, errors)

    def emit(self, record):
        try:
            if self.shouldRollover(record):
                with LockingTimedRotatingFileHandler._lock:
                    if self.shouldRollover(record):
                        self.doRollover()
            logging.FileHandler.emit(self, record)
        except Exception:
            self.handleError(record)

    def doRollover(self):
        """
        Практически в точности повторяет логику метода из родительского класса.
        Отличие от оригинального метода в том что если файл с именем baseFilename.rotation_time уже существует,
        то он не будет удаляться, т.к. это следствие того что другой процесс уже провел ротацию, вместо этого
        поток на файл просто переоткрывается, т.к. старый может указывать на уже перемещенный файл.
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        # get the time that this sequence started at and make it a TimeTuple
        currentTime = int(time.time())
        dstNow = time.localtime(currentTime)[-1]
        t = self.rolloverAt - self.interval
        if self.utc:
            timeTuple = time.gmtime(t)
        else:
            timeTuple = time.localtime(t)
            dstThen = timeTuple[-1]
            if dstNow != dstThen:
                if dstNow:
                    addend = 3600
                else:
                    addend = -3600
                timeTuple = time.localtime(t + addend)
        dfn = self.rotation_filename(self.baseFilename + "." +
                                     time.strftime(self.suffix, timeTuple))
        if os.path.exists(dfn):
            self.close()
            self.stream = self._open()
        else:
            self.rotate(self.baseFilename, dfn)
            if self.backupCount > 0:
                for s in self.getFilesToDelete():
                    os.remove(s)
            if not self.delay:
                self.stream = self._open()
        newRolloverAt = self.computeRollover(currentTime)
        while newRolloverAt <= currentTime:
            newRolloverAt = newRolloverAt + self.interval
        # If DST changes and midnight or weekly rollover, adjust for this.
        if (self.when == 'MIDNIGHT' or self.when.startswith('W')) and not self.utc:
            dstAtRollover = time.localtime(newRolloverAt)[-1]
            if dstNow != dstAtRollover:
                if not dstNow:  # DST kicks in before next rollover, so we need to deduct an hour
                    addend = -3600
                else:  # DST bows out before next rollover, so we need to add an hour
                    addend = 3600
                newRolloverAt += addend
        self.rolloverAt = newRolloverAt
